import os
import re
import math
import cv2
# import torch.backends.cudnn as cudnn
import gst_helper
import gst_admeta as admeta
import gst_cv_helper
from gi.repository import Gst, GObject, GLib, GstVideo
import numpy as np 
from timeit import default_timer as timer
import keypoints as kps
from pose import Pose, track_poses
from random import random as rand, randint as rint
import datetime
import time
import subprocess as sp
import logging
from logging.handlers import TimedRotatingFileHandler
import ftplib
import platform


#--- 宣告log檔格式
#log_file_name = 'log_' + datetime.datetime.now().strftime('%Y%m%d') + '.txt'
#logHandler = TimedRotatingFileHandler("/home/adlink/C6C_Airshower_Pose_Detect/log/" + log_file_name, when="midnight", interval=1) 
#logFormatter = logging.Formatter('%(asctime)s %(message)s') 
#logHandler.setFormatter(logFormatter) 
#logger = logging.getLogger('MyLogger') 
#logger.addHandler(logHandler)
#logger.setLevel(logging.INFO)
#---

kpt_names = ['nose', 'neck',
			 'r_sho', 'r_elb', 'r_wri', 'l_sho', 'l_elb', 'l_wri',
			 'r_hip', 'r_knee', 'r_ank', 'l_hip', 'l_knee', 'l_ank',
			 'r_eye', 'l_eye',
			 'r_ear', 'l_ear']

StartFlag = 0
EndFlag = 0
HandFlag = 0

def pad_width(h,w, stride, pad_value, min_dims):
	h = min(min_dims[0], h)
	min_dims[0] = math.ceil(min_dims[0] / float(stride)) * stride
	min_dims[1] = max(min_dims[1], w)
	min_dims[1] = math.ceil(min_dims[1] / float(stride)) * stride
	pad = []
	pad.append(int(math.floor((min_dims[0] - h) / 2.0)))
	pad.append(int(math.floor((min_dims[1] - w) / 2.0)))
	pad.append(int(min_dims[0] - h - pad[0]))
	pad.append(int(min_dims[1] - w - pad[1]))
	return pad

##### Analyze inference result context #####

def analyze_blob_size(val):
	try:
		mask_format = re.compile("\((?:\d+,)*\d+\)")
		mask = []
		for m in mask_format.findall(val.replace(" ", "")):
			mask.append(tuple(map(int, m.replace("(", "").replace(")", "").split(","))))
		return mask
	except:
		pass

	return


def analyze_mask(s):
	try:
		mask_format = re.compile("\((?:\d+,)*\d+\)")
		mask = []
		for m in mask_format.findall(s.replace(" ", "")):
			mask.append(tuple(map(int, m.replace("(", "").replace(")", "").split(","))))
		return mask
	except:
		pass

	return

num_keypoints = Pose.num_kpts
stride = 8
upsample_ratio = 4
track=False
pad_value=(0, 0, 0)


def decode(input_height,input_width,results, height, width, labels=[], fps=None):
	start = timer()
	heatmaps=results['heatmap']
	pafs=results['paf']
	heatmaps = np.transpose(heatmaps, (1, 2, 0))
	heatmaps = cv2.resize(heatmaps, (0, 0), fx=upsample_ratio, fy=upsample_ratio, interpolation=cv2.INTER_CUBIC)
	pafs = np.transpose(pafs, (1, 2, 0))
	pafs = cv2.resize(pafs, (0, 0), fx=upsample_ratio, fy=upsample_ratio, interpolation=cv2.INTER_CUBIC)
	total_keypoints_num = 0
	ttotal_keypoints_num = 0
	all_keypoints_by_type = []
	tall_keypoints_by_type = []
	scalew = input_width / width
	scaleh = input_height / height
	min_dims = [input_height, input_height]

	pad = pad_width(height, width, stride, pad_value, min_dims)
	
	total_keypoints_num = kps.fast_extract_keypoints(heatmaps[:, :, 0:Pose.num_kpts], all_keypoints_by_type)#,ttotal_keypoints_num)

	pose_entries, all_keypoints = kps.group_keypoints(all_keypoints_by_type, pafs, demo=False)
	
	for kpt_id in range(all_keypoints.shape[0]):
		all_keypoints[kpt_id, 0] = (all_keypoints[kpt_id, 0] * stride / upsample_ratio - pad[1]) / scalew
		all_keypoints[kpt_id, 1] = (all_keypoints[kpt_id, 1] * stride / upsample_ratio - pad[0]) / scaleh
	#print("Processed adyolac.py all_keypoints : %.1f ms" % (1000 * (timer() - te)))
	current_poses = []
	bdp = []
	#te =timer()
	for n in range(len(pose_entries)):
		if len(pose_entries[n]) == 0:
			continue
		pose_keypoints = np.ones((num_keypoints, 2), dtype=np.int32) * -1
		for kpt_id in range(num_keypoints):
			if pose_entries[n][kpt_id] != -1.0:  # keypoint was found
				pose_keypoints[kpt_id, 0] = int(all_keypoints[int(pose_entries[n][kpt_id]), 0])
				pose_keypoints[kpt_id, 1] = int(all_keypoints[int(pose_entries[n][kpt_id]), 1])
		pose = Pose(pose_keypoints, pose_entries[n][18])
		current_poses.append(pose)
	for pose in current_poses:
		bdp.append(pose.draw())
	return current_poses,bdp
	
############################################
StartLine_box = []
StartLine_box.append((640,200))
StartLine_box.append((1280,200))
StartLine_box.append((1280,900))
StartLine_box.append((640,900))
EndLine_box = []
EndLine_box.append((640,850))
EndLine_box.append((1280,850))
EndLine_box.append((1280,1000))
EndLine_box.append((640,1000))

def draw_boxs(img, boxs):
	h, w, c = img.shape
	face = cv2.FONT_HERSHEY_COMPLEX
	scale = 1.5
	thickness = 2
	baseline = 0
	for each in boxs:
		#l =  each.obj_label.decode("utf-8").strip() if each.obj_label.decode("utf-8").strip() != '' else str(each.class_id)
		#size = cv2.getTextSize(l, face, scale, thickness+1)
		x1, x2, y1, y2 = int(each.x1*w), int(each.x2*w), int(each.y1*h), int(each.y2*h)
		cv2.rectangle(img, (x1, y1), (x2, y2), (255, 255, 0), thickness)
		cv2.rectangle(img, (x1, y1), (x1+size[0][0], y1+size[0][1]+size[1]), (255, 255, 255), -1)
		#cv2.putText(img, l, (x1, y1 + size[0][1]), face, scale, (0, 0, 255), thickness+1)

def draw_area_boundary(img, roi_boundary, color):
	for i in range(len(roi_boundary)):
		try:
			cv2.line(img, tuple(roi_boundary[i]), tuple(roi_boundary[i+1]), color, 5)
		except:
			cv2.line(img, tuple(roi_boundary[i]), tuple(roi_boundary[0]), color, 5)
	return img
			
##### GStreamer python plugin context #####

def gst_video_caps_make(fmt):
	return  "video/x-raw, "\
		"format = (string) " + fmt + " , "\
		"width = " + GstVideo.VIDEO_SIZE_RANGE + ", "\
		"height = " + GstVideo.VIDEO_SIZE_RANGE + ", "\
		"framerate = " + GstVideo.VIDEO_FPS_RANGE


def arms_overhead(pose):    
    if pose.keypoints[4][1] == -1 and pose.keypoints[7][1] == -1 :
        return False
    if (pose.keypoints[7][1] == -1 and (pose.keypoints[2][1] < pose.keypoints[4][1]))or(pose.keypoints[4][1] == -1 and (pose.keypoints[5][1] < pose.keypoints[7][1] )):
        return True
    if (pose.keypoints[2][1] < pose.keypoints[4][1])or(pose.keypoints[5][1] < pose.keypoints[7][1]):
        return True
    
    #return all(y < pose.keypoints[1][1] for y in (pose.keypoints[3][1], pose.keypoints[4][1], pose.keypoints[6][1], pose.keypoints[7][1]))


def alarm_ring():
	sound_proc = sp.Popen('cvlc --play-and-exit /home/adlink/C6C_Airshower_Pose_Detect/arms_overhead.mp3', shell=True)
	#try:
	#	subprocess.communicate(timeout=5)
	#except subprocess.TimeoutExpired:
	#	sound_proc.kill()
	
	#print('sound_proc.poll() = ', sound_proc.poll() )
	#if sound_proc.poll() is not None:
	#	sound_proc.terminate()
	#	sound_proc.kill()
	#	self.sound_process.terminate()
		
	'''
    pygame.mixer.init()
    pygame.mixer.music.set_volume(1.0)
    pygame.mixer.music.load('alarm.mp3')
    pygame.mixer.music.play()
    
    #if not pygame.mixer.music.get_busy():
    #    #pygame.mixer.music.load('/home/adlink/human_pose_lite/alarm.mp3')
    #    pygame.mixer.music.load('alarm.mp3')
    #    pygame.mixer.music.play()	
	'''

#檢查FTP IP
FTP_IP_IS_alive = 'False'

def alarm_photo(area, img):
    log_path = '/home/adlink/C6C_Airshower_Pose_Detect/pic/'
    theTime = datetime.datetime.now()
    theTime = datetime.datetime.now().strftime('%Y%m%d%H%M%S%f')[:-1] #取得現在時間 時分秒毫秒作為檔名    
    file_name = area + '_Arms_Not_Overhead_' + theTime + '.jpg'
    #cv2.imwrite(os.path.join(log_path, file_name), img)
    cv2.imwrite(log_path + file_name, img)
    
    #global FTP_IP_IS_alive
    if FTP_IP_IS_alive == True:
        ftp_upload(file_name, os.path.join(log_path, file_name).replace('\\','/'))    
    
    return file_name


def ftp_upload(file_name, file_local):       
    host = '172.28.33.10'
    username = 'h10'
    password = 'cfi'
    f = ftplib.FTP(host)  # 實例化FTP對象
    f.login(username, password)  # 登錄
    
    #print("file_local:", file_local)
    #print("file_name:", file_name)
    
    #獲取當前路徑
    pwd_path = f.pwd()
    #print("FTP當前路徑:", pwd_path)
    
    #進FTP哪個資料夾
    f.cwd('\C6C_AI_FILES')
    pwd_path = f.pwd()    
        
    fp = open(file_local,'rb')    
    f.storbinary('STOR '+ file_name, fp)
    fp.close() 


def ipcheck(ip):
    ping_str = "-n 1" if  platform.system().lower()=="windows" else "-c 1"
    args = "ping " + " " + ping_str + " " + ip 
    status, result = sp.getstatusoutput(args) 
    
    if status == 0:
        #print("System " + str(ip) + " is UP !")
        return True
    else:
        #print("System " + str(ip) + " is DOWN !")
        return False


#檢查FTP IP        
if ipcheck('172.28.33.10') == True:
    FTP_IP_IS_alive = True
else:
    FTP_IP_IS_alive = False
    

class adtrans_openposePy(Gst.Element):
	GST_PLUGIN_NAME = 'adtrans_openpose_py'

	__gstmetadata__ = ("adtrans_openpose",
										 "GstElement",
										 "This is OpenPose Translator python.",
										 "Hiro Nguyen")

	__gsttemplates__ = (Gst.PadTemplate.new("src",
																					Gst.PadDirection.SRC,
																					Gst.PadPresence.ALWAYS,
																					Gst.Caps.from_string(gst_video_caps_make("{ BGR }"))),
											Gst.PadTemplate.new("sink",
																					Gst.PadDirection.SINK,
																					Gst.PadPresence.ALWAYS,
																					Gst.Caps.from_string(gst_video_caps_make("{ BGR }"))))

	_sinkpadtemplate = __gsttemplates__[1]
	_srcpadtemplate = __gsttemplates__[0]

	__gproperties__ = {
		"batch-num": (int, "batch-num", "Batch number",
									1, 256, 1, GObject.ParamFlags.READWRITE),
		"blob-size": (str, "blob-size", """inference output blob's output width and height, can using comma to define multiple output blob's size.""",
									 "(19,32,57),(38,32,57)", GObject.ParamFlags.READWRITE),
		"label-file": (str, "label-file", "Label file",
									 "", GObject.ParamFlags.READWRITE),
		"input-width": (int, "input-width", "Input width of inference",
										1, 65536, 456, GObject.ParamFlags.READWRITE),
		"input-height": (int, "input-height", "Input height of inference",
										 1, 65536, 256, GObject.ParamFlags.READWRITE),
		"threshold": (float, "threshold", "Threshold of box confidence",
										 0.0, 1.0, 0.8, GObject.ParamFlags.READWRITE),
	}

	def __init__(self):
		# Initialize properties before Base Class initialization
		self.batch_num, self.input_width, self.input_height, self.threshold = 1, 456, 256, 0.8
		self.blob_size_str = '(19,32,57),(38,32,57)'
		self.blob_size = analyze_blob_size(self.blob_size_str)
		self.label_file, self.labels = "", []

		super(adtrans_openposePy, self).__init__()

		self.sinkpad = Gst.Pad.new_from_template(self._sinkpadtemplate, 'sink')

		self.sinkpad.set_chain_function_full(self.chainfunc, None)

		self.sinkpad.set_chain_list_function_full(self.chainlistfunc, None)

		self.sinkpad.set_event_function_full(self.eventfunc, None)
		self.add_pad(self.sinkpad)

		self.srcpad = Gst.Pad.new_from_template(self._srcpadtemplate, 'src')

		self.srcpad.set_event_function_full(self.srceventfunc, None)

		self.srcpad.set_query_function_full(self.srcqueryfunc, None)
		self.add_pad(self.srcpad)

	def do_get_property(self, prop: GObject.GParamSpec):
		if prop.name == 'batch-num':
			return self.batch_num
		elif prop.name == 'blob-size':
			return self.blob_size_str
		elif prop.name == 'input-width':
			return self.input_width
		elif prop.name == 'input-height':
			return self.input_height
		elif prop.name == 'threshold':
			return self.threshold
		elif prop.name == 'label-file':
			return self.label_file
		else:
			raise AttributeError('unknown property %s' % prop.name)

	def do_set_property(self, prop: GObject.GParamSpec, value):
		if prop.name == 'batch-num':
			self.batch_num = int(value)
		elif prop.name == 'blob-size':
			blob_size = analyze_blob_size(str(value))
			if blob_size is None:
				raise AttributeError('Incurrect blob size format %s' % value)
			self.blob_size_str, self.blob_size = str(value), blob_size
		elif prop.name == 'input-width':
			self.input_width = int(value)
		elif prop.name == 'input-height':
			self.input_height = int(value)
		elif prop.name == 'threshold':
			self.threshold = float(value)
		elif prop.name == 'label-file':
			self.label_file = str(value)
			if os.path.isfile(self.label_file):
				with open(self.label_file, 'r') as f:
					self.labels = list(map(lambda l: l.strip(), f))
		else:
			raise AttributeError('unknown property %s' % prop.name)

	def chainfunc(self, pad: Gst.Pad, parent, buff: Gst.Buffer) -> Gst.FlowReturn:
          print('Adtranslator plugin only support BufferList, should be connected with inference plugin')
          return Gst.FlowReturn.ERROR

	def chainlistfunc(self, pad: Gst.Pad, parent, buff_list: Gst.BufferList) -> Gst.FlowReturn:
		#start = timer()
		#temp = timer()
		global HandFlag
		global StartFlag
		#global FTP_IP_IS_alive
		#檢查FTP IP        
        #if ipcheck('172.28.33.10') == True:
        #    FTP_IP_IS_alive = True
        #else:
        #    FTP_IP_IS_alive = False
		out_sizes = []

		for each in self.blob_size:
			sizes = 1
			for ea in each:
				sizes = sizes * int(ea)
			out_sizes.append(sizes)
		keys = ['heatmap','paf']
		blob_value = []
		with gst_helper.get_inference_data_to_numpy(buff_list, (sum(out_sizes)*2)) as data:
			offset = sum(out_sizes)
			for idx, s in enumerate(out_sizes):
				blob = data[offset:offset+s].reshape(self.blob_size[idx])
				blob_value.append(blob)
				offset += s
		net_outs =dict(zip(keys, blob_value))
		buf = gst_helper._gst_get_buffer_list_writable_buffer(buff_list, 0)
		img = gst_cv_helper.pad_and_buffer_to_numpy(pad, buf, ro=False)
		height, width, c = img.shape
		current_poses,bdp = decode(self.input_height,self.input_width,net_outs,height, width)
		arr = []

		for i, pose in enumerate(current_poses):
			if (pose.confidence > self.threshold ):
				# arr.append(admeta._DetectionBox(i, i, i, i,
				# 							  pose.bbox[0]/width,
				# 							  pose.bbox[1]/height,
				# 							  (pose.bbox[0] + pose.bbox[2])/width,
				# 							  (pose.bbox[1] + pose.bbox[3])/height,
				# 							  pose.confidence))
				for j, part in enumerate(bdp[i]):
					arr.append(admeta._DetectionBox(i, "", j, kpt_names[part[4]]+","+kpt_names[part[5]],
											  part[0]/width,
											  part[1]/height,
											  (part[2])/width,
											  (part[3])/height,
											  pose.confidence,"pose"))
			#detect pose
			#print('pose.keypoints[1][1] = ', pose.keypoints[1][1])
			#if pose.keypoints[1][1] > 520 and StartFlag == 0: #800 #if pose.keypoints[1][1] >= 520:
			StartFlag = 1
				#HandFlag = 0
				#print('Start Detect')
				#alarm_ring()

			#手臂過頭
			if arms_overhead(pose):
				cv2.rectangle(img, (pose.bbox[0]-100, pose.bbox[1] - 50), (pose.bbox[0]+430, pose.bbox[1]), (0, 0, 0), -1 )
				cv2.putText(img, 'arms overhead', (pose.bbox[0]-100, pose.bbox[1] - 10), cv2.FONT_HERSHEY_TRIPLEX, 2, (255, 255, 255), 1, cv2.LINE_AA)
				HandFlag = 1
				print('arms_overhead')
											
			# draw End line
			#draw_area_boundary(img, EndLine_box, (100,100,255)) 
			#if ank 10/13
#			if pose.keypoints[10][1] != -1 or pose.keypoints[13][1] != -1:
				#print('pose.keypoints[10][1] = ', pose.keypoints[10][1])
				#print('pose.keypoints[13][1] = ', pose.keypoints[13][1])
#				if pose.keypoints[10][1] <= 810 and pose.keypoints[13][1] <= 810: #850
				#EndFlag = 1 
#					if StartFlag == 1 :
#						if HandFlag == 0 :
#							print('Error')
							#alarm_ring()
#							cv2.putText(img, 'arms not overhead', (pose.bbox[0]-150, pose.bbox[1]), cv2.FONT_HERSHEY_TRIPLEX, 2, (0, 0, 255), 2, cv2.LINE_AA)
#							file_name = alarm_photo('Air_Shower', img)
#							logger.info('C6C_Air_Shower' + ' Arms_Not_Overhead ' + file_name)
#						else:
#							print('Pass')
#					StartFlag = 0
				#startFlag = 0
				#EndFlag = 0			
			
		#buf = gst_helper._gst_get_buffer_list_writable_buffer(list, 0)
		admeta.set_detection_box(buf, pad, arr)
		buff_list.remove(1, 1)
		#print("Processed adyolac.py : %.1f ms" % (1000 * (timer() - start)))
		return self.srcpad.push(buff_list.get(0))

	def eventfunc(self, pad: Gst.Pad, parent, event: Gst.Event) -> bool:
		return self.srcpad.push_event(event)

	def srcqueryfunc(self, pad: Gst.Pad, parent, query: Gst.Query) -> bool:
		return self.sinkpad.query(query)

	def srceventfunc(self, pad: Gst.Pad, parent, event: Gst.Event) -> bool:
		return self.sinkpad.push_event(event)


GObject.type_register(adtrans_openposePy)
__gstelementfactory__ = (adtrans_openposePy.GST_PLUGIN_NAME,
												 Gst.Rank.NONE, adtrans_openposePy)

###########################################

